#include <stdio.h>
#include <math.h>

#define PRIMES 20000

#ifndef TRUE
#define FALSE 0
#define TRUE !FALSE
#endif

int pn=8;
int p[PRIMES]={2,3,5,7,11,13,17,19};

int
is_prime(int n)
{
	int i;
	int l = sqrt(n);
	for(i=0;i<pn && p[i]<=n;i++) {
		if(n%p[i]==0)
			return FALSE;
	}
	return TRUE;
}

void
make_primes(void)
{
	int i;
	int start = 3;
	p[0]=2;
	p[1]=3;
	if(pn < 2)
		pn = 2;

	for(i=p[pn-1]+2;pn<PRIMES;i+=2) {
		if(is_prime(i)) {
			p[pn++]=i;
		}
	}
}

void
test_decreasing(void)
{
	int i;
	for(i=0;i<PRIMES-1;i++) {
		if(p[i]>p[i+1]) {
			printf("OVERFLOW at prime #%d\n",i);
			exit(1);
		}
	}
}

int
in_primes(int n)
{
	int i;
	for(i=0;i<PRIMES && n>=p[i];i++) {
		if(n==p[i])
			return 1;
	}
	return 0;
}

void
test_goldberg(void)
{
	int n;
	for(n=4;n<p[PRIMES-1];n+=2) {
		int i;
		int x,y;
		int got = FALSE;
		for(x=0;!got && x<PRIMES;x++) {
			for(y=0;y<PRIMES;y++) {
				if(p[x]+p[y]==n) {
					got = TRUE;
					break;
				}
			}
		}
		if(!got) {
			printf("Number %d is not a sum of two primes!\n");
			return;
		}
	}
	printf("Goldberg conjecture verified for numbers up to %d\n",n-2);
}

void
count_doubles(void)
{
	int i;
	int doubles=0;
	for(i=0;i<PRIMES-1;i++) {
		if(p[i]+2==p[i+1]) {
			doubles++;
		}
	}
	printf("%d double primes\n",doubles);
}

void
load_primes(void)
{
	FILE *fp = fopen("primes.dat","r");
	if(!fp)
		return;
	read(fileno(fp),&pn,sizeof(int));
	if(pn>PRIMES)
		pn = PRIMES;
	read(fileno(fp),p,sizeof(int)*pn);
	fclose(fp);
}

void
save_primes(void)
{
	FILE *fp = fopen("primes.dat","w");
	if(!fp)
		return;
	write(fileno(fp),&pn,sizeof(int));
	write(fileno(fp),p,sizeof(int)*pn);
	fclose(fp);
}
	

int
main(void)
{
	puts("Init");
	puts("Try to load primes");
	load_primes();
	puts("Make remaining primes");
	make_primes();
	puts("Save primes");
	save_primes();
	puts("Testing for overflows");
	test_decreasing();
	printf("Got a %d primes\n",PRIMES);
	printf("largest prime %d\n",p[PRIMES-1]);
	count_doubles();
	puts("Testing Goldberg conjectre");
	test_goldberg();
	return 0;
}
