#include <stdio.h>
#include <string.h>
#include <glib.h>

/* Turing Machine Mimimization by killing off unreachable parts

   It's based on turing.c, so read the comment in that file to figure out
   how the turing machine files look.

*/

/*Empty GList's for tape and rules, plus a pointer to the current
  GList node on the tape*/
static GList *rules = NULL;

/*Structure with info about 1 rule*/
typedef struct _Rule Rule;
struct _Rule {
	char *cur;	/*string with the current state*/
	char *next;	/*string with the next state for this rule*/
	char in;	/*what is the character that triggers this rule*/
	char out;	/*what is written back to tape*/
	char direction; /*direction as a character 'L' or 'R'*/
	gboolean to_keep; /*boolean signifies if this rule should be kept*/
};

/*reads the rules from the file and makes new Rule Structures for them and
  stuffs them into a (GList *)rules*/
static void
readrules(char *file)
{
	FILE *fp;
	Rule *r;
	char buf[256];
	
	/*read the turing machine from file*/
	if(!(fp = fopen(file,"r")))
		exit(1);
	/*always have one rule allocated*/
	r = g_new(Rule,1);
	r->to_keep = FALSE;
	while(fgets(buf,256,fp)) {
		char s1[256],s2[256];
		if(buf[0]=='#') /*signifies a comment*/
			continue;
		/*sscanf is braindead but works for simple inputs*/
		if(sscanf(buf,"%255s %255s %c %c %c",
			  s1,s2,&r->in,&r->out,&r->direction)==5) {
			r->cur = g_strdup(s1);
			r->next = g_strdup(s2);
			rules = g_list_append(rules,r);
			/*always have one rule allocated*/
			r = g_new(Rule,1);
			r->to_keep = FALSE;
		}
	}
	/*we free the rule which wasn't used*/
	g_free(r);
	fclose(fp);
}

/* mark state and all states that it can lead to */
static void
mark_state(char *state)
{
	GList *li;
	for(li=rules;li;li=li->next) {
		Rule *r = li->data;
		if(!r->to_keep && strcmp(r->cur,state)==0) {
			r->to_keep = TRUE;
			mark_state(r->next);
		}
	}
}

/*the main function*/
int
main(int argc, char *argv[])
{
	GList *li;

	/*read file, if no file is given use turing.in*/
	readrules(argc>1?argv[1]:"turing.in");

	/* mark all states starting with START */
	mark_state("START");

	/* print out the machine */
	for(li=rules;li!=NULL;li=li->next) {
		Rule *r = li->data;
		/* if we don't keep the rule, comment it out */
		if(!r->to_keep)
			printf("#");
		printf("%s %s %c %c %c\n",r->cur,r->next,r->in,r->out,r->direction);
	}

	return 0;
}
